/*
* Copyright (c) 1997 - 2016
* Actelion Pharmaceuticals Ltd.
* Gewerbestrasse 16
* CH-4123 Allschwil, Switzerland
*
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice, this
*    list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
* 3. Neither the name of the the copyright holder nor the
*    names of its contributors may be used to endorse or promote products
*    derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*/
 package com.actelion.research.gui.dnd;


import com.actelion.research.chem.StereoMolecule;
import com.actelion.research.chem.StructureInfo;
import com.actelion.research.chem.StructureTransferData;

import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;

public class SDFileMoleculeTransferable extends MoleculeTransferable
{
    public static final DataFlavor DF_SERIALIZEDSTRUCTURETRANSFERDATA =new DataFlavor(com.actelion.research.chem.StructureTransferData.class,"Actelion Structure Transfer Data Class");

    private DataFlavor FLAVORS[] = {DF_SERIALIZEDSTRUCTURETRANSFERDATA};
    StructureInfo  si_ = null;
    public SDFileMoleculeTransferable(StereoMolecule mol, StructureInfo  si)
    {
        super(mol);
        si_ = si;
    }

    public synchronized DataFlavor[] getTransferDataFlavors()
    {
        DataFlavor[] res = FLAVORS;
        DataFlavor[] d = super.getTransferDataFlavors();
        if (d != null) {
            res = new DataFlavor[d.length+1];
            System.arraycopy(d,0,res,1,d.length);
            res[0] = DF_SERIALIZEDSTRUCTURETRANSFERDATA;
        } 
        return res;
    }

    public boolean isDataFlavorSupported(DataFlavor flavor)
    {
        if (flavor.equals(DF_SERIALIZEDSTRUCTURETRANSFERDATA))
            return true;
        return super.isDataFlavorSupported(flavor);
    }

    public synchronized Object getTransferData(DataFlavor flavor) throws
            UnsupportedFlavorException,IOException
    {
        if(flavor.equals(DF_SERIALIZEDSTRUCTURETRANSFERDATA)) {
            return new StructureTransferData(mMol,si_);
        }

        return super.getTransferData(flavor);

    }

    public String toString()
    {
        return "SDFileMoleculeTransferable";
    }

    public void lostOwnership(Clipboard clipboard,Transferable contents)
    {
    }
}
