package com.actelion.research.calc.histogram;

/**
 * ConstantsHistogram
 * <p>Modest v. Korff</p>
 * <p>
 * Created by korffmo1 on 11.01.19.
 */
public class ConstantsHistogram {

    public static final double TINY_FACTOR = 0.01;
}
