body { font-family: Verdana, Arial, Helvetica, sans-serif; font-size: 12pt; color: #000000; text-decoration: none }
h1 { font-size: 18pt; font-weight : bold; color: #C06020; margin-top: 13px; margin-bottom: 10px; }
h2 { font-size: 16pt; font-weight: bold; color: #8080FF }
h3 { font-size: 13pt; font-style: italic; font-weight: bold; color: #3030AA }
td { font-size: 12pt }
.menu { font-weight: bold }
.keyword { font-style: italic }
a:link { color: #000099; font-weight: bold; text-decoration: none }
a:visited { color: #660066; font-weight: bold; text-decoration: none }
a:hover { color: #990000; font-weight: bold; text-decoration: none }
tt.regex { white-space: nowrap; background-color: #FFECE0; }
tt.match { white-space: nowrap; background-color: #E0ECFF; }
tt.string { white-space: nowrap; background-color: #D0FFD0; }
table.wspace { border-width:0; }
td.wspace { margin:0px 4px; padding:0px; border-width:0px; }
table.colored {
	border-width: 2px;
	border-spacing: 2px;
	border-style: solid;
	border-color: #FFFFA0;
	border-collapse: collapse;
	background-color: white;
	}
th.colored {
	border-width: 1px;
	padding: 2px;
	border-style: solid;
	border-color: #FFFFA0;
	background-color: #FFFFA0;
	}
td.colored {
	border-width: 1px;
	padding: 2px;
	border-style: solid;
	border-color: #FFFFA0;
	background-color: #FFFFF0;
	}
