package com.actelion.research.util.graph.complete;

/**
 * 
 * 
 * ICompleteGraph
 * <p>Copyright: Actelion Ltd., Inc. All Rights Reserved
 * This software is the proprietary information of Actelion Pharmaceuticals, Ltd.
 * Use is subject to license terms.</p>
 * @author Modest von Korff
 * @version 1.0
 * Oct 1, 2012 MvK: Start implementation
 */
public interface ICompleteGraph {
	
	/**
	 * 
	 * @return number of nodes in graph.
	 */
	public int getNumPPNodes();
}
