package com.actelion.research.chem.io.pdb.converter;

/**
 * ConstantsAminoAcidsLabeled
 * <p>Copyright: Idorsia Pharmaceuticals Ltd., Inc. All Rights Reserved
 * This software is the proprietary information of Idorsia Pharmaceuticals, Ltd.
 * Use is subject to license terms.</p>
 * Created by korffmo1 on 17.04.18.
 */
public class ConstantsAminoAcidsLabeled {

    public static final String TAG_AAS_NAME = "Name";

    public static final String TAG_AAS_ABBR1 = "Abbreviation1";

    public static final String RESOURCE_AAS_LABELED = "/resources/pdb/aminoAcidsLabeled.txt";

}
